#include "pngimg.h"
#include <stdio.h>
#include <assert.h>



/* 
 *
 * libpng 1.0.18 - December 3, 2004
 * For conditions of distribution and use, see copyright notice in png.h
 * Copyright (c) 1998-2004 Glenn Randers-Pehrson
 * (Version 0.96 Copyright (c) 1996, 1997 Andreas Dilger)
 * (Version 0.88 Copyright (c) 1995, 1996 Guy Eric Schalnat, Group 42, Inc.)
 *
 * This program reads in a PNG image, writes it out again, and then
 * compares the two files.  If the files are identical, this shows that
 * the basic chunk handling, filtering, and (de)compression code is working
 * properly.  It does not currently test all of the transforms, although
 * it probably should.
 *
 * The program will report "FAIL" in certain legitimate cases:
 * 1) when the compression level or filter selection method is changed.
 * 2) when the maximum IDAT size (PNG_ZBUF_SIZE in pngconf.h) is not 8192.
 * 3) unknown unsafe-to-copy ancillary chunks or unknown critical chunks
 *    exist in the input file.
 * 4) others not listed here...
 * In these cases, it is best to check with another tool such as "pngcheck"
 * to see what the differences between the two files are.
 *
 * If a filename is given on the command-line, then this file is used
 * for the input, rather than the default "pngtest.png".  This allows
 * testing a wide variety of files easily.  You can also test a number
 * of files at once by typing "pngtest -m file1.png file2.png ..."
 *
 * Copyright (c) 2006, vitas@matfyz.cz
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 * 
 *     * Redistributions of source code must retain the above copyright
 *       notice, this list of conditions and the following disclaimer.
 *     
 *     * Redistributions in binary form must reproduce the above
 *       copyright notice, this list of conditions and the following
 *       disclaimer in the documentation and/or other materials provided
 *       with the distribution.
 *       
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

/* avg 'img' to subimage at 0,0 + dx + dy */

void
img_avg(t_img *img, int dx, int dy)
{
	int xy, ix, iy, sx, sy;
	int ix_step, iy_step;
	int r, g, b;

	//assert((img->w % x) == 0 );
	//assert((img->h % y) == 0 );

	ix_step = img->w / dx;
	iy_step = img->h / dy;

	assert(ix_step > 0);
	assert(iy_step > 0);

	xy = ix_step * iy_step;

	for(sx = 0 ; sx+ix_step <= img->w; sx+=ix_step)
	for(sy = 0 ; sy+iy_step <= img->h; sy+=iy_step)
	{
		r = g = b = 0;
		for(ix=0; ix < ix_step; ix++)
		for(iy=0; iy < iy_step; iy++)
		{
			r += IMG_GETPIXEL_R(img, sx + ix, sy + iy);
			g += IMG_GETPIXEL_G(img, sx + ix, sy + iy);
			b += IMG_GETPIXEL_B(img, sx + ix, sy + iy);
		}
		IMG_PUTPIXEL(img, sx/ix_step, sy/iy_step, r / xy, g / xy, b / xy);
	}
} /* img_avg */

/*
 * img1 -- from x1, y1 to +w, +h
 * img2 -- from 0,0 to +w, +h
 */
double
img_dist(t_img *img1,  t_img *img2, int x1, int y1, int w, int h)
{
	int ix, iy;
	int r, g, b;
	double dist = 0;

	assert( img1->w >= w + x1);
	assert( img1->h >= h + y1);
	assert( img2->w >= w);
	assert( img2->h >= h);

	for(ix=0; ix < w; ix++)
	for(iy=0; iy < h; iy++)
	{
		r = IMG_GETPIXEL_R(img1, x1 + ix, y1 + iy);
		g = IMG_GETPIXEL_G(img1, x1 + ix, y1 + iy);
		b = IMG_GETPIXEL_B(img1, x1 + ix, y1 + iy);

		r -= IMG_GETPIXEL_R(img2, ix, iy);
		g -= IMG_GETPIXEL_G(img2, ix, iy);
		b -= IMG_GETPIXEL_B(img2, ix, iy);

		dist += r * r + g * g + b * b;
	}
	return dist;
} /* img_dist */

t_img *
img_new()
{
	t_img *ret;
	ret = (t_img*) malloc(sizeof(t_img));

	ret->w = 0;
	ret->h = 0;
	ret->rb = 0;
	ret->dep = 0;
	ret->col_type = 0;

	ret->data = NULL;
	ret->rows = NULL;

	return ret;
} /* img_new */

t_img*
img_create(int w, int h, int dep)
{
	t_img *img;
	int i;

	img = img_new();
	img->w = w;
	img->h = h;
	img->dep = dep;
	img->rb = w* dep;
	img->col_type = 0;

	img->data = (uch*) malloc(img->h * img->rb);

	if( NULL == img->data ) {
		printf("img_create(h%i w%i rb%i):7\n", img->h, img->w, img->rb);
		return NULL;
	}

	img->rows = (uch**) malloc(sizeof(uch*) * img->h );

	if( NULL == img->rows ) {
		printf("img_create:8\n");
		return NULL;
	}
	for(i = 0; i < img->h;  i++) {
		img->rows[i] = &img->data[i*img->rb];
	}
	return img;
}

t_img*
img_load(char *file_name)
{
	int interlace_type;
	int header;
	int i;
	png_structp png_ptr;
	png_infop info_ptr;
	png_infop end_info;
	int is_png;
	int png_transforms;
	t_img *img;

	FILE *fp = fopen(file_name, "rb");

	if (!fp) {
		printf("img_load:cannt fopen img '%i'\n", file_name);
		return NULL;
	}

	img = img_new();

	fread(&header, 1, sizeof(header), fp);
	is_png = !png_sig_cmp((png_bytep)&header, 0, sizeof(header));

	if (!is_png) {
		printf("img_load:'%s' is not png\n", file_name);
		return NULL;
	}

	png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
	if (!png_ptr) {
		printf("img_load:3\n");
		return NULL;
	}

	info_ptr = png_create_info_struct(png_ptr);

	if (!info_ptr) {
		png_destroy_read_struct(&png_ptr, (png_infopp)NULL, (png_infopp)NULL);
		printf("img_load:4\n");
		return NULL;
	}

	if (setjmp(png_jmpbuf(png_ptr)))
	{
		/* Free all of the memory associated with the png_ptr and info_ptr */
		png_destroy_read_struct(&png_ptr, &info_ptr, png_infopp_NULL);
		fclose(fp);
		/* If we get here, we had a problem reading the file */
		printf("img_load:5\n");
		return NULL;
	}

	end_info = png_create_info_struct(png_ptr);

	if (!end_info) {
		png_destroy_read_struct(&png_ptr, &info_ptr, (png_infopp)NULL);
		printf("img_load:6\n");
		return NULL;
	}

	png_init_io(png_ptr, fp);

	png_set_sig_bytes(png_ptr, sizeof(header));
	png_read_info(png_ptr, info_ptr);


	png_get_IHDR(png_ptr, info_ptr,
		&img->w, &img->h,
		&img->dep, &img->col_type,
		&interlace_type, NULL, NULL);
	png_set_strip_16(png_ptr);
	png_set_strip_alpha(png_ptr);

	if (img->col_type == PNG_COLOR_TYPE_GRAY && img->dep < 8)
		png_set_gray_1_2_4_to_8(png_ptr);
       
	png_transforms = PNG_TRANSFORM_PACKING;

	img->w = png_get_image_width(png_ptr, info_ptr);
	img->h = png_get_image_height(png_ptr, info_ptr);
	img->rb = png_get_rowbytes(png_ptr, info_ptr); 

	img->data = (uch*) malloc(img->h * img->rb);

	if( NULL == img->data ) {
		printf("img_load(h%i w%i rb%i):7\n", img->h, img->w, img->rb);
		return NULL;
	}

	img->rows = (uch**) malloc(sizeof(uch*) * img->h );

	if( NULL == img->rows ) {
		printf("img_load:8\n");
		return NULL;
	}
	for(i = 0; i < img->h;  i++) {
		img->rows[i] = &img->data[i*img->rb];
	}

	png_read_image(png_ptr, img->rows);

	//png_read_png(png_ptr, info_ptr, png_transforms, NULL);
	png_read_end(png_ptr, info_ptr);

	/* destroy all funny struct */
	png_destroy_read_struct(&png_ptr, &info_ptr, NULL);
	fclose(fp);

	return img;
} /* load_img */

void
img_save(t_img *img, char *filename)
{
	FILE *fp;
	png_structp png_ptr;
	png_infop info_ptr;


	fp = fopen(filename, "wb");
	if (fp == NULL) {
		printf("img_save:0\n");
		exit(-1);
	}

	png_ptr = png_create_write_struct(
		PNG_LIBPNG_VER_STRING,
		NULL, NULL, NULL);
		/// png_voidp user_error_ptr, user_error_fn, user_warning_fn);

	if (png_ptr == NULL)
	{
		fclose(fp);
		printf("img_save:1\n");
		exit(-1);
	}
	/* Allocate/initialize the image information data.  REQUIRED */
	info_ptr = png_create_info_struct(png_ptr);
	if (info_ptr == NULL)
	{
		fclose(fp);
		png_destroy_write_struct(&png_ptr,  png_infopp_NULL);
		printf("img_save:2\n");
		exit(-1);
	}


	if (setjmp(png_jmpbuf(png_ptr)))
	{
		/* If we get here, we had a problem reading the file */
		fclose(fp);
		png_destroy_write_struct(&png_ptr, &info_ptr);
		printf("img_save:3\n");
		exit(-1);
	}

	png_init_io(png_ptr, fp);

	png_set_IHDR(png_ptr, info_ptr,
	img->w, img->h, img->dep,
		PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
		PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_BASE);

	#if 0 
	sig_bit.red = true_red_bit_depth;
	sig_bit.green = true_green_bit_depth;
	sig_bit.blue = true_blue_bit_depth;
	png_set_sBIT(png_ptr, info_ptr, sig_bit);

	#endif
	png_write_info(png_ptr, info_ptr);
	png_write_image(png_ptr, img->rows);
	png_write_end(png_ptr, info_ptr);

	png_destroy_write_struct(&png_ptr, &info_ptr);

	fclose(fp);

}

/*
 * copy content img1 to img2 at possition x,y
 */
void
img_cpy(t_img *img1, t_img *img2, int x, int y)
{
	assert(img1->w+x <= img2->w);
	assert(img1->h+y <= img2->h);
	int i, j;


	for(i=0; i < img1->w; i++)
	for(j=0; j < img1->h; j++)
	{
		IMG_PUTPIXEL(img2, x +i, y+j,  
			IMG_GETPIXEL_R(img1, i, j),
			IMG_GETPIXEL_G(img1, i, j),
			IMG_GETPIXEL_B(img1, i, j));
	}


}


void
img_free(t_img* img)
{
	free(img->rows);
	free(img->data);
	free(img);
}

